#ifndef _SLIST_H_
#define _SLIST_H_

#include "common.h"

/*
 * Circular single-linked list implementation
 */
struct slist_head;

/**
 * Single-linked list structure
 */
struct slist_head {
	struct slist_head *head;	/*!< pointer to list head */
	struct slist_head *next;	/*!< next element (first) */
};


/**
 * slist_init_head
 * \brief	Initialize linked list
 * \param[in]	h	pointer to linked list head
 */
void slist_init_head( struct slist_head *h );

/**
 * slist_entry
 * \brief Get current element
 */
#define slist_entry(ptr, type, member) \
	container_of(ptr, type, member)


/**
 * slist_insert_after
 * \brief	Insert new element after given one
 * \param[in]	elem	Element to insert after
 * \param[in]	new	New element
 */
void slist_insert_after( struct slist_head *elem, struct slist_head *enew );

/**
 * slist_insert_before
 * \brief	Insert new element before given one
 * \param[in]	elem	Element to insert before
 * \param[in]	new	New element
 */
void slist_insert_before( struct slist_head *elem, struct slist_head *enew );


/**
 * slist_insert_head
 * \brief	Insert new element at first position
 * \param[in]	list	The list
 * \param[in]	new	New element
 */
#define slist_insert_head(list, enew) \
	slist_insert_after((list)->head, enew)

/**
 * slist_insert_tail
 * \brief	Insert new element at the end of the list
 * \param[in]	list	The list
 * \param[in]	new	New element
 */
void slist_insert_tail( struct slist_head *list, struct slist_head *enew );


/**
 * slist_remove
 * \brief	Remove element from the list
 * \param[in]	elem	Element to remove
 */
void slist_remove( struct slist_head *elem );


/**
 * slist_first
 * \brief	Get first element in the list
 * \param[in]	elem	Any list element
 * \return	First element
 */
struct slist_head * slist_first( struct slist_head *elem );


/**
 * slist_last
 * \brief	Get last element in the list
 * \param[in]	elem	Any list element
 * \return	Last element
 */
struct slist_head * slist_last( struct slist_head *elem );


/**
 * slist_next
 * \brief	Get next element after given one
 * \param[in]	elem	Any list element
 * \return	Next element
 */
struct slist_head * slist_next( struct slist_head *elem );


/**
 * slist_is_head
 * \brief	Check if the element passed is the list head
 * \param[in]	elem	Any list element
 * \return	0 if the element passed is not the head, 1 otherwise
 */
int slist_is_head( struct slist_head *elem );


/**
 * slist_is_first
 * \brief	Check if the element passed is the first one
 * \param[in]	elem	Any list element
 * \return	0 if the element passed is not the first, 1 otherwise
 */
int slist_is_first( struct slist_head *elem );


/**
 * slist_is_last
 * \brief	Check if the element passed is the last one
 * \param[in]	elem	Any list element
 * \return	0 if the element passed is not the last, 1 otherwise
 */
int slist_is_last( struct slist_head *elem );


/**
 * slist_replace
 * \brief	Replace an element in the list with a new one
 * \param[in]	old	Element to replace
 * \param[in]	new	Replacement
 */
void slist_replace( struct slist_head *old, struct slist_head *enew );


/**
 * slist_for_each
 * \brief	Walk the list of elements in a for like loop
 * \param[in]	head	The list
 * \param[in]	p	Pointer to store the current element
 * \note	Not slist_remove() safe!
 */
#define slist_for_each(head, p)		\
	for (p = (typeof(*p) *)(head)->next;	\
	p != (typeof(*p) *)(head);		\
	p = (typeof(*p) *)(p)->next)


/**
 * slist_for_each_safe
 * \brief	Walk the list of elements in a for like loop
 * \param[in]	head	The list
 * \param[in]	n	Temporary storage
 * \param[in]	p	Pointer to store the current element
 * \note	slist_remove() safe version
 */
#define slist_for_each_safe(head, n, p)	\
	for (p = (typeof(*p) *)(head)->next, n = (typeof(*p) *)(p)->next; \
	p != (typeof(*p) *)(head); 		\
	p = (typeof(*p) *)(n), n = (typeof(*p) *)(p)->next)


/**
 * slist_is_empty
 * \brief	Check if the list has no elements
 * \param[in]	head	The list
 * \return	0 if the list is not empty, 1 otherwise
 */
#define slist_is_empty(head) ({		\
	((head)->next == head);		\
	})


/**
 * slist_num_entries
 * \brief	Get the number of elements in the list
 * \param[in]	head	The list
 * \return	number of entries
 */
#define slist_num_entries(head) ({		\
	int _x = 0;				\
	struct slist_head *p = (head);		\
	for (; p && p->next && p->next != (head); _x++) p = p->next;	\
	_x; })


#endif
